import { EditorView, ViewPlugin, logException } from '@codemirror/next/view';
import { Facet, combineConfig, Precedence, StateEffect, StateField, Transaction } from '@codemirror/next/state';
import { keymap } from '@codemirror/next/keymap';
import { tooltips, showTooltip } from '@codemirror/next/tooltip';

/// Denotes how to
/// [filter](#autocomplete.autocomplete^config.filterType)
/// completions.
var FilterType;
(function (FilterType) {
    /// Only show completions that start with the currently typed text.
    FilterType[FilterType["Start"] = 0] = "Start";
    /// Show completions that have the typed text anywhere in their
    /// content.
    FilterType[FilterType["Fuzzy"] = 1] = "Fuzzy";
})(FilterType || (FilterType = {}));
class AutocompleteContext {
    /// @internal
    constructor(explicit, filterType) {
        this.explicit = explicit;
        this.filterType = filterType;
    }
    filter(completion, text) {
        if (this.filterType == FilterType.Start)
            return completion.length > text.length && completion.slice(0, text.length) == text;
        else
            return completion.length > text.length && completion.indexOf(text) > -1;
    }
}
function retrieveCompletions(state, pos, context) {
    let found = state.languageDataAt("autocomplete", pos);
    function next(i) {
        if (i == found.length)
            return Promise.resolve([]);
        return Promise.resolve(found[i](state, pos, context)).then(result => result.length ? result : next(i + 1));
    }
    return next(0);
}
const autocompleteConfig = Facet.define({
    combine(configs) {
        return combineConfig(configs, {
            override: null,
            filterType: FilterType.Start
        });
    }
});
/// Returns an extension that enables autocompletion.
function autocomplete(config = {}) {
    return [
        activeCompletion,
        autocompleteConfig.of(config),
        autocompletePlugin,
        style,
        tooltips(),
        Precedence.Override.set(keymap([
            { key: "ArrowDown", run: moveCompletion("down") },
            { key: "ArrowUp", run: moveCompletion("up") },
            { key: "Enter", run: acceptCompletion },
            { key: "Escape", run: closeCompletion }
        ]))
    ];
}
function moveCompletion(dir) {
    return (view) => {
        let active = view.state.field(activeCompletion);
        if (!(active instanceof ActiveCompletion))
            return false;
        let selected = (active.selected + (dir == "up" ? active.options.length - 1 : 1)) % active.options.length;
        view.dispatch(view.state.update({ effects: selectCompletion.of(selected) }));
        return true;
    };
}
function acceptCompletion(view) {
    let active = view.state.field(activeCompletion);
    if (!(active instanceof ActiveCompletion))
        return false;
    applyCompletion(view, active.options[active.selected]);
    return true;
}
/// Explicitly start autocompletion.
function startCompletion(view) {
    let active = view.state.field(activeCompletion);
    if (active != null && active != "pending")
        return false;
    view.dispatch(view.state.update({ effects: toggleCompletion.of(true) }));
    return true;
}
function applyCompletion(view, option) {
    let apply = option.apply || option.label;
    // FIXME make sure option.start/end still point at the current
    // doc, or keep a mapping in an active completion
    if (typeof apply == "string") {
        view.dispatch(view.state.update({
            changes: { from: option.start, to: option.end, insert: apply },
            selection: { anchor: option.start + apply.length }
        }));
    }
    else {
        apply(view);
    }
}
function closeCompletion(view) {
    let active = view.state.field(activeCompletion);
    if (active == null)
        return false;
    view.dispatch(view.state.update({ effects: toggleCompletion.of(false) }));
    return true;
}
const openCompletion = StateEffect.define();
const toggleCompletion = StateEffect.define();
const selectCompletion = StateEffect.define();
const activeCompletion = StateField.define({
    create() { return null; },
    update(value, tr) {
        let event = tr.annotation(Transaction.userEvent);
        if (event == "input" || event == "delete" && value)
            value = "pending";
        else if (tr.docChanged || tr.selection)
            value = null;
        for (let effect of tr.effects) {
            if (effect.is(openCompletion))
                value = new ActiveCompletion(effect.value, 0);
            else if (effect.is(toggleCompletion))
                value = effect.value ? "pendingExplicit" : null;
            else if (effect.is(selectCompletion) && value instanceof ActiveCompletion)
                value = new ActiveCompletion(value.options, effect.value, value.id, value.tooltip);
        }
        return value;
    },
    provide: [
        showTooltip.nFrom(active => active instanceof ActiveCompletion ? active.tooltip : none),
        EditorView.contentAttributes.from(active => active instanceof ActiveCompletion ? active.attrs : baseAttrs)
    ]
});
const baseAttrs = { "aria-autocomplete": "list" }, none = [];
class ActiveCompletion {
    constructor(options, selected, id = "cm-ac-" + Math.floor(Math.random() * 1679616).toString(36), tooltip = [completionTooltip()]) {
        this.options = options;
        this.selected = selected;
        this.id = id;
        this.tooltip = tooltip;
        this.attrs = {
            "aria-autocomplete": "list",
            "aria-activedescendant": this.id + "-" + this.selected,
            "aria-owns": this.id
        };
    }
}
function createListBox(completion) {
    const ul = document.createElement("ul");
    ul.id = completion.id;
    ul.setAttribute("role", "listbox");
    ul.setAttribute("aria-expanded", "true");
    for (let i = 0; i < completion.options.length; i++) {
        const li = ul.appendChild(document.createElement("li"));
        li.id = completion.id + "-" + i;
        li.innerText = completion.options[i].label;
        li.setAttribute("role", "option");
    }
    return ul;
}
// We allocate a new function instance every time the completion
// changes to force redrawing/repositioning of the tooltip
function completionTooltip() {
    return (view) => {
        let active = view.state.field(activeCompletion);
        let list = createListBox(active);
        list.addEventListener("click", (e) => {
            let index = 0, dom = e.target;
            for (;;) {
                dom = dom.previousSibling;
                if (!dom)
                    break;
                index++;
            }
            let active = view.state.field(activeCompletion);
            if (active instanceof ActiveCompletion && index < active.options.length)
                applyCompletion(view, active.options[index]);
        });
        function updateSel(view) {
            let cur = view.state.field(activeCompletion);
            if (cur instanceof ActiveCompletion)
                updateSelectedOption(list, cur.selected);
        }
        return {
            dom: list,
            mount: updateSel,
            update(update) {
                if (update.state.field(activeCompletion) != update.prevState.field(activeCompletion))
                    updateSel(update.view);
            },
            pos: active.options.reduce((m, o) => Math.min(m, o.start), 1e9),
            style: "autocomplete"
        };
    };
}
function updateSelectedOption(list, selected) {
    let set = null;
    for (let opt = list.firstChild, i = 0; opt; opt = opt.nextSibling, i++) {
        if (i == selected) {
            if (!opt.hasAttribute("aria-selected")) {
                opt.setAttribute("aria-selected", "true");
                set = opt;
            }
        }
        else {
            if (opt.hasAttribute("aria-selected"))
                opt.removeAttribute("aria-selected");
        }
    }
    if (set)
        scrollIntoView(list, set);
}
function scrollIntoView(container, element) {
    let parent = container.getBoundingClientRect();
    let self = element.getBoundingClientRect();
    if (self.top < parent.top)
        container.scrollTop -= parent.top - self.top;
    else if (self.bottom > parent.bottom)
        container.scrollTop += self.bottom - parent.bottom;
}
const DebounceTime = 100;
const autocompletePlugin = ViewPlugin.fromClass(class {
    constructor(view) {
        this.view = view;
        this.stateVersion = 0;
        this.debounce = -1;
    }
    update(update) {
        if (!update.docChanged && !update.selectionSet &&
            update.prevState.field(activeCompletion) == update.state.field(activeCompletion))
            return;
        this.stateVersion++;
        if (this.debounce > -1)
            clearTimeout(this.debounce);
        let active = update.state.field(activeCompletion);
        this.debounce = active == "pending" || active == "pendingExplicit"
            ? setTimeout(() => this.startUpdate(active == "pendingExplicit"), DebounceTime) : -1;
    }
    startUpdate(explicit) {
        this.debounce = -1;
        let version = this.stateVersion, state = this.view.state, pos = state.selection.primary.head;
        let config = state.facet(autocompleteConfig);
        let context = new AutocompleteContext(explicit, config.filterType);
        (config.override ? Promise.resolve(config.override(state, pos, context)) : retrieveCompletions(state, pos, context))
            .then(result => {
            if (this.stateVersion != version || result.length == 0)
                return;
            this.view.dispatch(this.view.state.update({ effects: openCompletion.of(result) }));
        })
            .catch(e => logException(this.view.state, e));
    }
});
const style = EditorView.baseTheme({
    "tooltip.autocomplete": {
        fontFamily: "monospace",
        overflowY: "auto",
        maxHeight: "10em",
        listStyle: "none",
        margin: 0,
        padding: 0,
        "& > li": {
            cursor: "pointer",
            padding: "1px 1em 1px 3px",
            lineHeight: 1.2
        },
        "& > li[aria-selected]": {
            background_fallback: "#bdf",
            backgroundColor: "Highlight",
            color_fallback: "white",
            color: "HighlightText"
        }
    }
});

export { AutocompleteContext, FilterType, autocomplete, startCompletion };
